/*
 * AIEngine a new generation network intrusion detection system.
 *
 * Copyright (C) 2013-2022  Luis Campo Giralte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com>
 *
 */
#include "HTTPSession.h"
#include <boost/spirit/include/qi.hpp> // For parsing the 5 tuple
#if defined(PYTHON_BINDING)
#include <boost/python/stl_iterator.hpp>
#endif

namespace qi = boost::spirit::qi;

namespace aiengine {

/* LCOV_EXCL_START */

SystemPtr system = SystemPtr(new System());

void HTTPSession::start() {

	do_read();
}

void HTTPSession::do_read() {

        // Make the request empty before reading,
        // otherwise the operation behavior is undefined.
        request = {};

        bhttp::async_read(socket_, buffer_, request,
        	std::bind(
                	&HTTPSession::handle_read,
                        shared_from_this(),
                        std::placeholders::_1,
                        std::placeholders::_2));
}

void HTTPSession::handle_read(boost::system::error_code ec, std::size_t bytes_transferred) {

        boost::ignore_unused(bytes_transferred);

        total_bytes_ += bytes_transferred;

        // This means the client closed the connection
        if (ec)
                return do_close();

        ++ total_requests_;
        last_time_ = std::time(nullptr);
        process_request(std::move(request));
}

void HTTPSession::handle_write( boost::system::error_code ec, std::size_t bytes_transferred, bool close) {

        boost::ignore_unused(bytes_transferred);

        total_bytes_ += bytes_transferred;
        ++ total_responses_;

        if (ec) {
                // std::cout << "Error on write:" << ec.message() << "\n";
                return;
        }

        if (close) {
                // This means we should close the connection, usually because
                // the response indicated the "Connection: close" semantic.
                return do_close();
        }

        // We're done with the response so delete it
        res_ = nullptr;

        // Read another request
        do_read();
}

void HTTPSession::do_close() {

	socket_.close();
}

void HTTPSession::process_request(bhttp::request<bhttp::dynamic_body>&& req) {

        bhttp::response<bhttp::dynamic_body> response;
        response.version(req.version());
        response.keep_alive(req.keep_alive());
        response.set(bhttp::field::server, "AIEngine " VERSION);
        response.content_length(0);

	if (stack_ == nullptr) {
		std::string_view message = "No stack available";

		response.keep_alive(false);
		response.set(bhttp::field::content_type, "text/html");
		response.result(bhttp::status::service_unavailable);
		boost::beast::ostream(response.body()) << message;
		response.content_length(message.length());

		AIINFO << "[" << socket_.remote_endpoint().address().to_string() << "] " << req.method()
			<< " " << req.target() << " " << response[bhttp::field::content_type]
			<< " " << response.result_int();

		return send_response(std::move(response));
	}

	if (req.method() == bhttp::verb::get)
        	handle_get_message(response);
	else if (req.method() == bhttp::verb::post)
        	handle_post_message(response);
	else if (req.method() == bhttp::verb::put)
        	handle_put_message(response);
	else {
		response.result(bhttp::status::bad_request);
                response.set(bhttp::field::content_type, "text/plain");
                boost::beast::ostream(response.body())
                	<< "Invalid request-method '"
                        << request.method_string().to_string()
                        << "'";
        }

        AIINFO << "[" << socket_.remote_endpoint().address().to_string() << "] " << req.method()
		<< " " << req.target() << " " << response[bhttp::field::content_type]
		<< " " << response.result_int();

	return send_response(std::move(response));
}

void HTTPSession::handle_get_message(bhttp::response<bhttp::dynamic_body> &response) {

	std::string uri(request.target());

	if (uri.compare(0, strlen(http_uri_show_protocols_summary), http_uri_show_protocols_summary) == 0)
		show_protocols_summary(response);
	else if (uri.compare(0, strlen(http_uri_show_network_flows), http_uri_show_network_flows) == 0)
		show_network_flows(response);
	else if (uri.compare(0, strlen(http_uri_show_protocol), http_uri_show_protocol) == 0)
		show_protocol_summary(response);
	else if (uri.compare(0, strlen(http_uri_show_summary), http_uri_show_summary) == 0)
		show_summary(response);
	else if (uri.compare(0, strlen(http_uri_show_system), http_uri_show_system) == 0)
		show_system(response);
	else if (uri.compare(0, strlen(http_uri_pcapfile), http_uri_pcapfile) == 0)
		show_pcapfile(response);
	else if (uri.compare(0, strlen(http_uri_show_network_flow), http_uri_show_network_flow) == 0)
		show_network_flow(response);
	else if (uri.compare(0, strlen(http_uri_show_current_packet), http_uri_show_current_packet) == 0)
		show_current_packet(response);
	else if (uri.compare(0, strlen(http_uri_show_logs), http_uri_show_logs) == 0)
		show_logs(response);
	else if (uri.compare(0, strlen(http_uri_show_health), http_uri_show_health) == 0)
		show_health(response);
#if defined(PYTHON_BINDING)
	else if (uri.compare(0, strlen(http_uri_show_anomalies), http_uri_show_anomalies) == 0)
		show_anomalies(response);
	else if (uri.compare(0, strlen(http_uri_python_script), http_uri_python_script) == 0)
		show_python_script(response);
	else if (uri.compare(0, strlen(http_uri_show_python_locals), http_uri_show_python_locals) == 0)
		show_python_locals(response);
	else if (uri.compare(0, strlen(http_uri_show_python_help), http_uri_show_python_help) == 0)
		show_python_help(response);
#endif
	else
		show_uris(response);
}

void HTTPSession::handle_post_message(bhttp::response<bhttp::dynamic_body> &response) {

        std::string uri(request.target());

        if (uri.compare(0, strlen(http_uri_pcapfile), http_uri_pcapfile) == 0)
                upload_pcapfile(response);
#if defined(PYTHON_BINDING)
        else if (uri.compare(0, strlen(http_uri_python_script), http_uri_python_script) == 0)
                upload_python_script(response);
#endif
	else {
		response.result(bhttp::status::bad_request);
		response.set(bhttp::field::content_type, "text/plain");
		boost::beast::ostream(response.body()) << "Invalid POST resource '" << uri << "'";
		response.content_length(response.payload_size());
	}
}

void HTTPSession::show_protocols_summary(bhttp::response<bhttp::dynamic_body> &response) {

	std::ostringstream out;

        response.set(bhttp::field::content_type, "text/plain");

	if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0) {
        	if (std::size_t found = accept.find("application/json"); found != std::string::npos) {
                	Json j;
                        response.set(bhttp::field::content_type, "application/json");
			stack_->showProtocolSummary(j);
                        out << j;
                } else
			stack_->showProtocolSummary(out);
	} else
		stack_->showProtocolSummary(out);

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

// The network flows uri looks like
// /v1/flows
// /v1/flows?limit=100
// /v1/flows?l7protocol=dns&limit=100
// /v1/flows?l7protocol=ssl

void HTTPSession::show_network_flows(bhttp::response<bhttp::dynamic_body> &response) {

	std::ostringstream out;
	std::string uri(request.target());
	bool jsonize = false;
	UserFlowOptions options;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
			jsonize = true;

	uri.erase(0, strlen(http_uri_show_network_flows));

	// Check if the URI contains parameters
	if (uri.find("?") != std::string::npos) {
		std::vector<std::string> parameters;

		uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

		boost::split(parameters, uri, boost::is_any_of("&"));

		for (auto &parameter: parameters) {
			std::vector<std::string> entry;

			boost::split(entry, parameter, boost::is_any_of("="));

			if ((entry.size() == 2) and (entry[1].length() > 0)) {

				if (entry[0].compare("src_ip") == 0) {
					options.portsrc = std::atoi(entry[1].c_str());
					continue;
				}

				if (entry[0].compare("dst_port") == 0) {
					options.portdst = std::atoi(entry[1].c_str());
					continue;
				}

				if (entry[0].compare("src_ip") == 0) {
					options.ipsrc = entry[1];
					continue;
				}

				if (entry[0].compare("dst_ip") == 0) {
					options.ipdst = entry[1];
					continue;
				}

				if (entry[0].compare("protocol") == 0) {
					options.protocol = std::atoi(entry[1].c_str());
					continue;
				}

				if (entry[0].compare("limit") == 0) {
					options.limit = std::atoi(entry[1].c_str());
					continue;
				}

				if (entry[0].compare("l7protocol") == 0) {
					options.l7protocol_name = entry[1];
					continue;
				}
                        }
		}
	}

	if (jsonize) {
        	Json j;
                response.set(bhttp::field::content_type, "application/json");

		stack_->showFlows(j, options);

                out << j;
	} else {
        	response.set(bhttp::field::content_type, "text/plain");

		stack_->showFlows(out, options);
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

// The protocol uri looks like
// /v1/protocol?name=ssl
// /v1/protocol?name=ssl&stats_level=1
// /v1/protocol?name=http&map=host

void HTTPSession::show_protocol_summary(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        std::vector<std::string> parameters;
        std::string uri(request.target());
        int level = 1;
        std::string protocol = "";
	std::string map_name = "";
        bool jsonize = false;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        uri.erase(0, strlen(http_uri_show_protocol));

        // Check if the URI contains parameters
        if (uri.find("?") != std::string::npos) {
                std::vector<std::string> parameters;

                uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

                boost::split(parameters, uri, boost::is_any_of("&"));

                for (auto &parameter: parameters) {
                        std::vector<std::string> entry;

                        boost::split(entry, parameter, boost::is_any_of("="));

                        if ((entry.size() == 2) and (entry[1].length() > 0)) {

                                if (entry[0].compare("stats_level") == 0) {
                                        level = std::atoi(entry[1].c_str());
                                        continue;
                                }

                                if (entry[0].compare("map") == 0) {
                                        map_name = entry[1];
                                        continue;
                                }

                                if (entry[0].compare("name") == 0) {
                                        protocol = entry[1];
                                        continue;
                                }
			}
		}
	}

	if (protocol.length() > 0) { // The user setup the parameter name
		if (jsonize) {
			Json j;
			response.set(bhttp::field::content_type, "application/json");

			if (map_name.length() > 0)
				stack_->statistics(j, protocol, map_name);
			else
				stack_->statistics(j, protocol, level);

			out << j;
		} else {
			// Check if protocol is set to something
			response.set(bhttp::field::content_type, "text/plain");
			stack_->statistics(out, protocol, level);
		}
		response.result(bhttp::status::ok);
	} else {
		if (!jsonize) {
			response.set(bhttp::field::content_type, "text/plain");
			out << "No protocol name set";
		}
		response.result(bhttp::status::bad_request);
	}
	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

// /v1/summary

void HTTPSession::show_summary(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        bool jsonize = false;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        if (jsonize) {
                Json j;
                response.set(bhttp::field::content_type, "application/json");

		pdis_->statistics(j);

                out << j;
        } else {
                response.set(bhttp::field::content_type, "text/plain");
		pdis_->statistics(out);
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

// /v1/system
void HTTPSession::show_system(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        bool jsonize = false;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        if (jsonize) {
                Json j;
                response.set(bhttp::field::content_type, "application/json");

		system->statistics(j);

                out << j;
        } else {
                response.set(bhttp::field::content_type, "text/plain");
		system->statistics(out);
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

void HTTPSession::show_uris(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        bool jsonize = false;
	std::ostringstream baseuri;

	baseuri << "http://";

	if (boost::beast::string_view hostname = request[bhttp::field::host]; hostname.length() > 0)
		baseuri << hostname;
	else
		baseuri << socket_.local_endpoint().address().to_string() << ":" << socket_.local_endpoint().port();

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        if (jsonize) {
                nlohmann::json j;
                response.set(bhttp::field::content_type, "application/json");

		j["uris"] = {
			{ std::string(baseuri.str() + http_uri_show_protocols_summary), "Protocols summary" },
			{ std::string(baseuri.str() + http_uri_show_protocol), "Protocol summary" },
			{ std::string(baseuri.str() + http_uri_show_network_flows), "Network flows" },
			{ std::string(baseuri.str() + http_uri_show_summary), "Summary" },
			{ std::string(baseuri.str() + http_uri_show_system), "System" },
			{ std::string(baseuri.str() + http_uri_pcapfile), "Upload pcapfile" },
#if defined(PYTHON_BINDING)
			{ std::string(baseuri.str() + http_uri_python_script), "Upload python code" },
			{ std::string(baseuri.str() + http_uri_show_python_locals), "Python locals" },
			{ std::string(baseuri.str() + http_uri_show_anomalies), "Anomalies" },
#endif
			{ std::string(baseuri.str() + http_uri_show_network_flow), "Network flow" },
			{ std::string(baseuri.str() + http_uri_show_current_packet), "Current packet" },
			{ std::string(baseuri.str() + http_uri_show_logs), "Logs" },
			{ std::string(baseuri.str() + http_uri_show_health), "Health" }
		};

                out << j;
        } else {
                response.set(bhttp::field::content_type, "text/html");

		out << "<html><head><title>"
			<< "AIEngine " VERSION " running on " << system->getOSName()
			<< " kernel " << system->getReleaseName()
			<< " " << system->getVersionName()
			<< " " << system->getMachineName()
			<< " hostname " << system->getHostName()
			<< "</title></head>\n<body><center>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_protocols_summary << "\">Protocols summary</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_protocol << "\">Protocol summary</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_network_flows << "\">Network flows</a><br>\n"
#if defined(PYTHON_BINDING)
			<< "<a href=\"" << baseuri.str() << http_uri_show_anomalies << "\">Anomalies</a><br>\n"
#endif
			<< "<a href=\"" << baseuri.str() << http_uri_show_summary << "\">Summary</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_system << "\">System</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_pcapfile << "\">Upload pcapfile</a><br>\n"
#if defined(PYTHON_BINDING)
			<< "<a href=\"" << baseuri.str() << http_uri_python_script << "\">Python code</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_python_locals << "\">Python locals</a><br>\n"
#endif
			<< "<a href=\"" << baseuri.str() << http_uri_show_network_flow << "\">Network flow</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_current_packet << "\">Current packet</a><br>\n"
			<< "<a href=\"" << baseuri.str() << http_uri_show_logs << "\">Logs</a><br>\n"
#if defined(PYTHON_BINDING)
			<< "<a href=\"" << baseuri.str() << http_uri_show_python_help << "\">Help</a><br>\n"
#endif
			<< "</center></body>\n</html>\n";
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

void HTTPSession::show_pcapfile(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream html;

        response.set(bhttp::field::content_type, "text/html");

        html << "<html><head><title>Upload pcapfile</title></head>\n<body><center>\n"
		<< "<form id=\"uploadbanner\" enctype=\"multipart/form-data\" method=\"post\" >\n"
		<< " <input id=\"fileupload\" name=\"myfile\" type=\"file\" />\n"
		<< "<input type=\"submit\" value=\"submit\" id=\"submit\" />\n"
		<< "</form>\n"
		<< "</center></body>\n</html>\n";

	boost::beast::ostream(response.body()) << html.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

#if defined(PYTHON_BINDING)

void HTTPSession::show_python_script(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream html;

        response.set(bhttp::field::content_type, "text/html");

        html << "<html><head><title>Python code</title></head>\n<body><center>\n"
		<< "<form id=\"uploadbanner\" enctype=\"multipart/form-data\" method=\"post\" >\n"
		<< "<label>Python code</label><br>\n"
		<< "<textarea type=\"text\" id=\"code\" name=\"code\" rows=\"32\" cols=\"64\">\n"
		<< "</textarea><br>\n"
		<< "<input type=\"submit\" value=\"submit\" id=\"submit\" />\n"
		<< "</form>\n"
		<< "</center></body>\n</html>\n";

	boost::beast::ostream(response.body()) << html.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

boost::python::object dir(boost::python::object object) {
	boost::python::handle<> handle(PyObject_Dir(object.ptr()));
	return boost::python::object(handle);
}

bool callable(boost::python::object object) {
	return 1 == PyCallable_Check(object.ptr());
}

void HTTPSession::show_python_locals(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        bool jsonize = false;
        std::string uri(request.target());
	std::map<std::string, std::string> items;
	std::string object_name = "";

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

	// Check if there is parameters on the URI
        uri.erase(0, strlen(http_uri_show_python_locals));

        // Check if the URI contains parameters
        if (uri.find("?") != std::string::npos) {
                std::vector<std::string> parameters;
		std::string name = "";
		std::string property_name = "";

                uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

                boost::split(parameters, uri, boost::is_any_of("&"));

                for (auto &parameter: parameters) {
                        std::vector<std::string> entry;

                        boost::split(entry, parameter, boost::is_any_of("="));

                        if ((entry.size() == 2) and (entry[1].length() > 0)) {

                                if (entry[0].compare("name") == 0) {
                                        name = entry[1];
                                        continue;
                                }
                                if (entry[0].compare("property") == 0) {
                                        property_name = entry[1];
                                        continue;
                                }
                        }
                }
		if (name.length() > 0) {
			boost::python::dict locals(boost::python::borrowed(PyEval_GetLocals()));
			response.result(bhttp::status::not_found);

			// Search on locals if we have that object
			boost::python::list keys = boost::python::list(locals.keys());
			bool found = false;
			for (int i = 0; i < len(keys); ++i) {
				boost::python::extract<std::string> extractor(keys[i]);
				if (extractor.check()) {
					std::string key = extractor();

					if (key.compare(name) == 0) {
						found = true;
						break;
					}
				}
			}
			if (found) { // The object exists on locals
				object_name = name;
				boost::python::extract<boost::python::object> extractor(locals[name]);
				boost::python::object self = extractor();
				boost::python::extract<std::string> extract(self.attr("__class__").attr("__name__"));

				if (extract.check()) {
					std::string object_type = extract();
					typedef boost::python::stl_input_iterator<boost::python::str> iterator_type;
  					for (iterator_type name(dir(self)), end; name != end; ++name) { // Loop over the properties
						boost::python::extract<std::string> extract(*name);
						if (extract.check()) {
							std::string attribute_name = extract();
							if (property_name.length() > 0) {
								if (property_name.compare(attribute_name) == 0) {
									boost::python::object prop = self.attr(attribute_name.c_str());
									items[attribute_name] = boost::python::extract<std::string>(boost::python::str(prop))();
									break;
								}
							} else {
								if ((attribute_name.compare(0, 2, "__") != 0) and !callable(self.attr(*name))) {
									boost::python::object prop = self.attr(attribute_name.c_str());
									items[attribute_name] = boost::python::extract<std::string>(boost::python::str(prop))();
								}
							}
						}
					}
				}
				response.result(bhttp::status::ok);
			}
		}
        } else { // There is no parameters so just show the locals
		boost::python::dict locals(boost::python::borrowed(PyEval_GetLocals()));
		boost::python::list keys = boost::python::list(locals.keys());

		for (int i = 0; i < len(keys); ++i) {
			boost::python::extract<std::string> extractor(keys[i]);
			if (extractor.check()) {
				std::string key = extractor();

				boost::python::extract<boost::python::object> objectExtractor(locals[key]);
				boost::python::object obj = objectExtractor();
				std::string object_classname = boost::python::extract<std::string>(obj.attr("__class__").attr("__name__"));

				if (key.rfind("__", 0) != 0)
					items[key] = object_classname;
			}
		}
        	response.result(bhttp::status::ok);
	}

	if (jsonize) {
                nlohmann::json j;
                response.set(bhttp::field::content_type, "application/json");

		if (object_name.length() > 0)
			j["properties"] = items;
		else
			j["objects"] = items;

		out << j;
	} else {
        	response.set(bhttp::field::content_type, "text/plain");
		if (object_name.length() > 0) {
			out << "Object " << object_name << "\n";

			for (auto &item: items)
				out << "\t" << item.first << " = " << item.second << "\n";
		} else {
			out << "Python objects" << "\n";

			for (auto &item: items)
				out << "\t" << item.first << ": " << item.second << "\n";
		}
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

void HTTPSession::show_python_help(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        std::ostringstream filename;
        std::string code("import pydoc\npydoc.help(pyaiengine)");

        filename << "/tmp/aiengine.pcode." << getpid();

        bool executed = run_python_code(filename.str(), code);

        if (executed) {
		std::ifstream t(filename.str());

		out << t.rdbuf();

		t.close();

		response.result(bhttp::status::ok);
        } else {
                response.result(bhttp::status::bad_request);
        }

        response.set(bhttp::field::content_type, "text/plain");
	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

#endif

// Now POST methods

void HTTPSession::upload_pcapfile(bhttp::response<bhttp::dynamic_body> &response) {

        std::size_t off, roff, foff_s, foff_e;
        boost::beast::string_view ct = request[bhttp::field::content_type];
	response.set(bhttp::field::content_type, "text/plain");
        response.content_length(0);

#if defined(PYTHON_BINDING)
	if (!is_authorized(response))
		return;
#endif
        if (off = ct.find("multipart/form-data"); off == std::string::npos) {
                response.result(bhttp::status::unsupported_media_type);
                return;
        }

        if (off = ct.find("boundary="); off == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        // Now get the filename and the content type to verify that is a pcap file

        std::ostringstream body;
#if BOOST_BEAST_VERSION >= 266
        body << boost::beast::make_printable(request.body().data());
#else
        body << boost::beast::buffers(request.body().data());
#endif
        std::string boundary(ct.substr(off + 9));

        if (off = body.str().find(boundary); off == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        off += boundary.length() + 2;

        if (off = body.str().find("\r\n\r\n", off); off == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        if (roff = body.str().rfind(boundary); roff == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        // extract the file name
        if (foff_s = body.str().find("filename=", boundary.length() + 4); foff_s == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        if (foff_e = body.str().find("\r\n", foff_s); foff_e == std::string::npos) {
                response.result(bhttp::status::bad_request);
                return;
        }

        std::string filename(body.str().substr(foff_s + 10, foff_e - (foff_s + 10) - 1));
        std::ostringstream path;

        path << "/tmp/" << filename;
        std::ofstream ofs (path.str(), std::ofstream::out);

        int len = body.str().length() - (off + 4 + boundary.length() + 2 + 6);
        ofs << body.str().substr(off + 4, len);

        ofs.close();

        // Close the other device or pcapfiles
        pdis_->close();

        // Open the new one
        pdis_->open(path.str());

        // Run it
        pdis_->run();

        std::ostringstream redirect, html;

        redirect << "http://";

        if (boost::beast::string_view hostname = request[bhttp::field::host]; hostname.length() > 0)
                redirect << hostname;
        else
                redirect << socket_.local_endpoint().address().to_string() << ":" << socket_.local_endpoint().port();

        redirect << http_uri_show_uris;

        response.set(bhttp::field::content_type, "text/html");

	html << "<html><body>\n"
		<< "<script>\n"
		<< "setTimeout(function() {\n"
		<< "  window.location.href = '" << redirect.str() << "';\n"
		<< "}, 5000);\n"
		<< "</script>\n"
		<< "<p>Pcap upload successfully.</p>\n"
		<< "</body></html>";

	boost::beast::ostream(response.body()) << html.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

#if defined(PYTHON_BINDING)

bool HTTPSession::run_python_code(const std::string &filename, const std::string &code) const {

	bool executed = false;
	std::ostringstream scode;
	int pid = getpid();
#if defined(IS_DARWIN)
	boost::system::error_code ec;

	boost::filesystem::resize_file(filename, 0, ec); // truncate file
#else
	std::error_code ec;

	std::filesystem::resize_file(filename, 0, ec); // truncate file
#endif
	// We create a execution with a redirection of the output
	// bear in mind that the user could put print statements
	// so we need to redirect properly the things
	scode << "import sys\n"
		<< "output_" << pid << " = open('" << filename << "', 'w')\n"
		<< "temp_" << pid << " = sys.stdout\n"
		<< "sys.stdout = output_" << pid << "\n"
		<< code << "\n"
		<< "output_" << pid << ".close()\n"
		<< "sys.stdout = temp_" << pid << "\n"
		<< "del temp_" << pid << "\n"
		<< "del output_" << pid << "\n";

	std::ofstream term(filename, std::ios_base::out);
	auto &oldout = OutputManager::getInstance()->out();
	OutputManager::getInstance()->setOutput(term);

        try {
		[[maybe_unused]] PyGilContext gil_lock;

		// Retrieve the main module.
		boost::python::object main = boost::python::import("__main__");
		boost::python::object global(main.attr("__dict__"));

		boost::python::exec(scode.str().c_str(), global, global);
		executed = true;
        } catch (boost::python::error_already_set const &) {
		PyObject *ptype, *pvalue, *ptraceback;
		PyErr_Fetch(&ptype, &pvalue, &ptraceback);
		PyObject* str_exc_type = PyObject_Repr(pvalue); //Now a unicode object
		PyObject* pyStr = PyUnicode_AsEncodedString(str_exc_type, "utf-8", "Error ~");

		term << PyBytes_AS_STRING(pyStr) << std::endl;

		Py_XDECREF(str_exc_type);
		Py_XDECREF(pyStr);
        }

	term.close();
	OutputManager::getInstance()->setOutput(oldout);

	return executed;
}

// This function makes the Authorization of the requests
// Bear in mind that browsers will do a base64(username:password)
// so the value that needs to be stored should contain the : character
bool HTTPSession::is_authorized(bhttp::response<bhttp::dynamic_body> &response) const {

	std::string token(pdis_->getRawAuthorizationCodeToken());

	// Check if we have an authorization code token set on the PacketDispatcher
	if (token.length() > 0) {
		if (boost::beast::string_view authorization = request[bhttp::field::authorization]; authorization.length() > 0) {
			std::size_t found = authorization.find("Basic ");

			if (found != std::string::npos) {
				int token_length = authorization.length() - 6;
				if (authorization.compare(6, token_length, token) != 0) {
					boost::beast::ostream(response.body()) << "Wrong credentials";
					response.content_length(response.payload_size());
					response.result(bhttp::status::forbidden);
					return false;
				}
				return true; // The token is correct
			}
		}
		boost::beast::ostream(response.body()) << "Wrong authorization credentials";
		response.content_length(response.payload_size());
		response.set(bhttp::field::www_authenticate, "Basic realm='AIEngine'");
		response.result(bhttp::status::unauthorized);
		return false;
	}
	return true;
}

void HTTPSession::upload_python_script(bhttp::response<bhttp::dynamic_body> &response) {

        std::size_t off, roff;
        boost::beast::string_view ct = request[bhttp::field::content_type];
	std::string code;
        std::ostringstream body;
#if BOOST_BEAST_VERSION >= 266
        body << boost::beast::make_printable(request.body().data());
#else
        body << boost::beast::buffers(request.body().data());
#endif
	response.set(bhttp::field::content_type, "text/plain");
        response.content_length(0);

	if (!is_authorized(response))
		return;

        if (off = ct.find("multipart/form-data"); off != std::string::npos) {
		if (off = ct.find("boundary="); off == std::string::npos) {
                	response.result(bhttp::status::bad_request);
                	return;
        	}
        	std::string boundary(ct.substr(off + 9));

        	if (off = body.str().find(boundary); off == std::string::npos) {
                	response.result(bhttp::status::bad_request);
                	return;
        	}

		// Just jump to the next line and dont process Content-disposition
		if (off = body.str().find("\r\n\r\n", off); off == std::string::npos) {
                	response.result(bhttp::status::bad_request);
                	return;
		}

        	if (roff = body.str().rfind(boundary); roff == std::string::npos) {
                	response.result(bhttp::status::bad_request);
                	return;
        	}

		code = body.str().substr(off, roff - off - 4);

        } else if (off = ct.find("text/python"); off != std::string::npos) {
		code = body.str();
	} else {
                response.result(bhttp::status::unsupported_media_type);
                return;
	}

	if (code.length() > 0) {
		std::ostringstream filename, message, redirect, html;
		int delay = 0;

		response.set(bhttp::field::content_type, "text/html");
		filename << "/tmp/aiengine.pcode." << getpid();

		bool executed = run_python_code(filename.str(), code);

		std::ifstream t(filename.str());

		if (executed) {
			message << "Python code loaded successfully\n";
			delay = 5000;
			response.result(bhttp::status::ok);
		} else {
			message << "Python code can not be loaded\n";
			delay = 3000;
			response.result(bhttp::status::bad_request);
		}

		redirect << "http://";

		if (boost::beast::string_view hostname = request[bhttp::field::host]; hostname.length() > 0)
			redirect << hostname;
		else
			redirect << socket_.local_endpoint().address().to_string() << ":" << socket_.local_endpoint().port();

		redirect << http_uri_show_uris;

		html << "<html><body>\n"
			<< "<script>\n"
			<< "setTimeout(function() {\n"
			<< "  window.location.href = '" << redirect.str() << "';\n"
			<< "}, " << delay << ");\n"
			<< "</script>\n"
			<< "<p>" << message.str() << "</p>\n"
			<< "<pre>" << t.rdbuf() << "</pre>\n"
			<< "</body></html>";

		t.close();

		boost::beast::ostream(response.body()) << html.str();
		response.content_length(response.payload_size());
        } else {
		response.result(bhttp::status::bad_request);
		boost::beast::ostream(response.body()) << "No python code provided";
		response.content_length(response.payload_size());
	}
}

void HTTPSession::set_python_object_property(bhttp::response<bhttp::dynamic_body> &response) {

	std::ostringstream out;
	std::string uri(request.target());
	std::string name = "";
	std::string property_name = "";
	std::string property_value = "";

        // Check if there is parameters on the URI
        uri.erase(0, strlen(http_uri_show_python_locals));

        // Check if the URI contains parameters
        if (uri.find("?") != std::string::npos) {
                std::vector<std::string> parameters;

                uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

                boost::split(parameters, uri, boost::is_any_of("&"));

                for (auto &parameter: parameters) {
                        std::vector<std::string> entry;

                        boost::split(entry, parameter, boost::is_any_of("="));

                        if ((entry.size() == 2) and (entry[1].length() > 0)) {

                                if (entry[0].compare("name") == 0) {
                                        name = entry[1];
                                        continue;
                                }
                                if (entry[0].compare("property") == 0) {
                                        property_name = entry[1];
                                        continue;
                                }
                                if (entry[0].compare("value") == 0) {
                                        property_value = entry[1];
                                        continue;
                                }
                        }
                }
	}
	if ((name.length() > 0)and(property_name.length() > 0)and(property_value.length() > 0)) {
		std::ostringstream code;
		std::ostringstream filename;

		code << name << "." << property_name << " = " << property_value;

		filename << "/tmp/aiengine.pcode." << getpid();

		bool executed = run_python_code(filename.str(), code.str());

		std::ifstream t(filename.str());

		out << t.rdbuf();

		t.close();

		if (executed) {
			std::ostringstream redirect;

			redirect << "http://";

			if (boost::beast::string_view hostname = request[bhttp::field::host]; hostname.length() > 0)
				redirect << hostname;
			else
				redirect << socket_.local_endpoint().address().to_string() << ":" << socket_.local_endpoint().port();

			redirect << http_uri_show_uris;

			response.set(bhttp::field::location, redirect.str());
			response.result(bhttp::status::ok);
		} else
			response.result(bhttp::status::bad_request);
	} else {
		out << "Can not change the property";
		response.result(bhttp::status::not_found);
	}
	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

// /v1/anomalies
void HTTPSession::show_anomalies(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        bool jsonize = false;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        if (jsonize) {
                Json j;
                response.set(bhttp::field::content_type, "application/json");

                stack_->getAnomalyManager()->statistics(j);

                out << j;
        } else {
                response.set(bhttp::field::content_type, "text/plain");

                stack_->getAnomalyManager()->statistics(out);
        }

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
        response.result(bhttp::status::ok);
}

#endif

std::string HTTPSession::get_flowid_from_uri(const std::string &value) {
        std::string flowid = "";
	std::string uri = value;

        // Check if the URI contains a flow id parameters
        if (uri.find("?") != std::string::npos) {
                std::vector<std::string> parameters;

                uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

                boost::split(parameters, uri, boost::is_any_of("&"));

                for (auto &parameter: parameters) {
                        std::vector<std::string> entry;

                        boost::split(entry, parameter, boost::is_any_of("="));

                        if ((entry.size() == 2) and (entry[1].length() > 0)) {

                                if (entry[0].compare("id") == 0) {
                                        flowid = entry[1];
                                        continue;
                                }
                        }
                }
        }
	return flowid;
}

SharedPointer<Flow> HTTPSession::get_flow_from_uri(const std::string &flowid) const {

	SharedPointer<Flow> flow = nullptr;
	std::string address = flowid;
        std::string addrsrc, addrdst;
        int portsrc, portdst, proto;

	// The URI can have encoded charaters for [] so need to replace them
	boost::replace_all(address, "%5B", "[");
	boost::replace_all(address, "%5D", "]");

        // Do a basic parsing
        qi::phrase_parse(address.begin(), address.end(),
                qi::lit("[") >> *(~qi::char_("") - "]") >> qi::lit("]:") >>
                qi::int_ >> qi::lit(":[") >> *(~qi::char_("") - "]"),
                qi::ascii::space,
                addrsrc, proto, addrdst);

        std::size_t found = addrsrc.find_last_of(":");

        std::string ipsrc = addrsrc.substr(0, found);
        portsrc = std::atoi(std::string(addrsrc.substr(found + 1)).c_str());

        found = addrdst.find_last_of(":");

        std::string ipdst = addrdst.substr(0, found);
        portdst = std::atoi(std::string(addrdst.substr(found + 1)).c_str());

        if ((ipsrc.length() > 0)and(portsrc > 0)and(ipdst.length() > 0)and(portdst > 0))
                flow = stack_->getFlow(ipsrc.c_str(), portsrc, proto, ipdst.c_str(), portdst);

	return flow;
}

// URIS:
// /v1/flow?id=[10.42.0.42:42140]:6:[172.217.7.14:443]

void HTTPSession::show_network_flow(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        std::string uri(request.target());
        bool jsonize = false;
        std::string flowid;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        uri.erase(0, strlen(http_uri_show_network_flow));

	flowid = get_flowid_from_uri(uri);

	if (SharedPointer<Flow> flow = get_flow_from_uri(flowid); flow) {
        	response.result(bhttp::status::ok);
		std::time_t duration = std::time(nullptr) - flow->getArriveTime();
		std::time_t last_packet_seen = std::time(nullptr) - flow->getLastPacketTime();
		int timeout = stack_->getFlowsTimeout();;
		std::string_view status = "active";

		if (last_packet_seen > timeout) {
			if (last_packet_seen > timeout * 2)
				status = "comatose";
			else
				status = "timeout";
		}

		if (jsonize) {
			Json j;
			response.set(bhttp::field::content_type, "application/json");

			flow->show(j);

			j["status"] = status;
			j["duration"] = (int)duration;
			j["last_packet_seen"] = (int)last_packet_seen;
			out << j;
		} else {
			response.set(bhttp::field::content_type, "text/plain");
			std::string_view ev = flow->haveEvidence() ? "yes": "no";
			std::string_view reject = flow->isReject() ? "yes": "no";
 
			out << "Network flow:" << *flow << "\n"
				<< "Protocol:" << flow->getL7ProtocolName() << "\n"
				<< "Evidence:" << ev << "\n"
				<< "Reject:" << reject << "\n"
				<< "Status:" << status << "\n"
				<< "Up bytes:" << flow->total_bytes[static_cast<int>(FlowDirection::FORWARD)] << "\n"
				<< "Down bytes:" << flow->total_bytes[static_cast<int>(FlowDirection::BACKWARD)] << "\n"
				<< "Up packets:" << flow->total_packets[static_cast<int>(FlowDirection::FORWARD)] << "\n"
				<< "Down packets:" << flow->total_packets[static_cast<int>(FlowDirection::BACKWARD)] << "\n"
				<< "Duration:" << (int)duration << " secs\n"
				<< "Last packet seen:" << (int)last_packet_seen << " secs\n";

			flow->show(out);
		}
	} else {
		if (!jsonize) {
			if (flowid.length() > 0)
				out << "Network flow:" << flowid << " not found";
			else
				out << "Network flow: not set";
		}
                response.result(bhttp::status::not_found);
	}

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

void HTTPSession::handle_put_message(bhttp::response<bhttp::dynamic_body> &response) {

        std::string uri(request.target());

        if (uri.compare(0, strlen(http_uri_show_network_flow), http_uri_show_network_flow) == 0)
                update_network_flow(response);
#if defined(PYTHON_BINDING)
        else if (uri.compare(0, strlen(http_uri_show_python_locals), http_uri_show_python_locals) == 0)
      		set_python_object_property(response);
#endif
	else {
		response.result(bhttp::status::bad_request);
		response.set(bhttp::field::content_type, "text/plain");
		boost::beast::ostream(response.body()) << "Invalid PUT resource '" << uri << "'";
		response.content_length(response.payload_size());
	}
}

// Just json format is allowed
void HTTPSession::update_network_flow(bhttp::response<bhttp::dynamic_body> &response) {

        boost::beast::string_view content_type = request[bhttp::field::content_type];

        if (content_type.compare("application/json") == 0) {
		std::string uri(request.target());
                std::ostringstream bodydata;

#if BOOST_BEAST_VERSION >= 266
        	bodydata << boost::beast::make_printable(request.body().data());
#else
        	bodydata << boost::beast::buffers(request.body().data());
#endif
        	uri.erase(0, strlen(http_uri_show_network_flow));

		std::string flowid = get_flowid_from_uri(uri);

        	if (SharedPointer<Flow> flow = get_flow_from_uri(flowid); flow) {
                	response.result(bhttp::status::ok);

                	nlohmann::json j = nlohmann::json::parse(bodydata.str(), nullptr, false);
                	if (j.is_discarded()) {
                        	std::string_view message = "Not valid json content";

                        	response.set(bhttp::field::content_type, "text/html");
                        	response.result(bhttp::status::bad_request);
                        	boost::beast::ostream(response.body()) << message;
                        	response.content_length(message.length());
                	} else {
				// The user wants to change the label of the flow
				if (j.find("label") != j.end(); j["label"].is_string())
					flow->setLabel(j["label"]);

				// The user wants to have evidences of the flow
				if (j.find("evidence") != j.end(); j["evidence"].is_boolean())
					flow->setEvidence(j["evidence"]);

				// The user wants to reject the flow if supported
				if (j.find("reject") != j.end(); j["reject"].is_boolean())
					flow->setReject(j["reject"]);
#if defined(PYTHON_BINDING)
				// The user wants to detach the flow
				if (j.find("detach") != j.end(); j["detach"].is_boolean())
					if (j["detach"] == true)
						flow->detach();
#endif
			}
		} else
                	response.result(bhttp::status::not_found);
	} else
		response.result(bhttp::status::unsupported_media_type);
}

void HTTPSession::show_current_packet(bhttp::response<bhttp::dynamic_body> &response) {

        std::ostringstream out;
        std::vector<std::string> items;
        std::string uri(request.target());
        bool jsonize = false;

        if (boost::beast::string_view accept = request[bhttp::field::accept]; accept.length() > 0)
                if (std::size_t found = accept.find("application/json"); found != std::string::npos)
                        jsonize = true;

        if (jsonize) {
		Json j;
                response.set(bhttp::field::content_type, "application/json");

		pdis_->showCurrentPayloadPacket(j);
		out << j;
	} else {
		response.set(bhttp::field::content_type, "text/plain");

		pdis_->showCurrentPayloadPacket(out);
        }

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

void HTTPSession::show_logs(bhttp::response<bhttp::dynamic_body> &response) {

	std::ostringstream out;
	std::string uri(request.target());
	int limit = 16;
	glob_t globlist;
	const char *pattern_log_file = "aiengine-*.log";

        response.result(bhttp::status::ok);
        response.set(bhttp::field::content_type, "text/plain");

	// Check if there is parameters on the URI
        uri.erase(0, strlen(http_uri_show_logs));

        // Check if the URI contains parameters
        if (uri.find("?") != std::string::npos) {
                std::vector<std::string> parameters;
		std::string name = "";
		std::string property_name = "";

                uri.erase(std::remove(uri.begin(), uri.end(), '?'), uri.end());

                boost::split(parameters, uri, boost::is_any_of("&"));

                for (auto &parameter: parameters) {
                        std::vector<std::string> entry;

                        boost::split(entry, parameter, boost::is_any_of("="));

                        if ((entry.size() == 2) and (entry[1].length() > 0)) {

                                if (entry[0].compare("limit") == 0) {
                                        limit = std::atoi(entry[1].c_str());
                                        continue;
                                }
			}
		}
	}

      if (glob(pattern_log_file, 0, NULL, &globlist) == 0) {
		std::string filename = globlist.gl_pathv[globlist.gl_pathc - 1];

		FILE *fp = fopen(filename.c_str(), "r");
		if (fp) {
			fseek(fp, 0, SEEK_END);
			char buffer[512];
			auto pos = ftell(fp);
			int count = 0;

			while (pos) {
				if (!fseek(fp, --pos, SEEK_SET)) {
					if (fgetc(fp) == '\n')
						if (count++ == limit)
							break;
				}
			}
			if (count <= limit)
				fseek(fp, 0, SEEK_SET);

			while (fgets(buffer, sizeof(buffer), fp))
				out << buffer;

			fclose(fp);
		} else
			response.result(bhttp::status::not_found);
	} else
		response.result(bhttp::status::not_found);

	boost::beast::ostream(response.body()) << out.str();
	response.content_length(response.payload_size());
}

void HTTPSession::show_health(bhttp::response<bhttp::dynamic_body> &response) {

	// Just respond 200 that all is good :)
	boost::beast::ostream(response.body()) << "Up and running ;)";
	response.content_length(response.payload_size());
	response.set(bhttp::field::content_type, "text/plain");
        response.result(bhttp::status::ok);
}

/* LCOV_EXCL_STOP */

} // namespace aiengine
